package convertquery

import (
	"encoding/json"
	"errors"
	"fmt"
	"query-service/internal/entity"
)

/*
ConvertQuery converts a json string to an AQL query

Parameters: jsonMsg is the JSON file directly outputted by the drag and drop query builder in the frontend

Return: a string containing the corresponding AQL query, a string containing the database name and an error */
func (s *Service) ConvertQuery(jsonMsg *[]byte) (*string, *string, error) {

	jsonStruct, err := convertJSONToStruct(jsonMsg)
	if err != nil {
		fmt.Println(err)
		return nil, nil, err
	}

	// Check to make sure all indexes exist
	// How many entities are there
	numEntities := len(jsonStruct.Entities) - 1
	// How many relations there are
	numRelations := len(jsonStruct.Relations) - 1

	// Make sure no entity should be returned that is outside the range of that list
	for _, e := range jsonStruct.Return.Entities {
		// If this entity references an entity that is outside the range
		if e > numEntities || e < 0 {
			return nil, nil, errors.New("non-existing entity referenced in return")
		}
	}

	// Make sure that no relation mentions a non-existing entity
	for _, r := range jsonStruct.Relations {
		if r.EntityFrom > numEntities || r.EntityTo > numEntities {
			return nil, nil, errors.New("non-exisiting entity referenced in relation")
		}
	}

	// Make sure no non-existing relation is tried to be returned
	for _, r := range jsonStruct.Return.Relations {
		if r > numRelations || r < 0 {
			return nil, nil, errors.New("non-existing relation referenced in return")
		}
	}

	result := createQuery(jsonStruct)
	return result, &jsonStruct.DatabaseName, nil
}

/* convertJSONtoStruct reads a JSON file and sorts the data into the appropriate structs
Parameters: jsonMsg is the JSON file directly outputted by the drag and drop query builder in the frontend

Return: parsedJSON is a struct with the same structure and holding the same data as jsonMsg
*/
func convertJSONToStruct(jsonMsg *[]byte) (*entity.QueryParsedJSON, error) {
	jsonStruct := entity.QueryParsedJSON{}
	err := json.Unmarshal(*jsonMsg, &jsonStruct)

	if err != nil {
		return nil, err
	}

	return &jsonStruct, nil
}

/* createQuery generates a query based on the json file provided
Parameters: jsonQuery is a parsedJSON struct holding all the data needed to form a query

Return: a string containing the corresponding AQL query and an error
*/
func createQuery(jsonQuery *entity.QueryParsedJSON) *string {
	// GROTE SIDENOTE:
	// Vrij zeker dat een query waar alléén edges worden opgevraagd (#4)
	// niet wordt gesupport door zowel de result parser als de frontend receiver

	var (
		relationsToReturn []string
		nodesToReturn     []string
		nodeUnion         string
		relationUnion     string
	)

	// Loop over all relations
	ret := ""

	for i, relation := range jsonQuery.Relations {
		relationName := fmt.Sprintf("r%v", i)

		if relation.EntityFrom >= 0 {
			// if there is a from-node
			// create the let for this node
			fromName := fmt.Sprintf("n%v", relation.EntityFrom)

			ret += *createNodeLet(&jsonQuery.Entities[relation.EntityFrom], &fromName)

			ret += *createRelationLetWithFromEntity(&relation, relationName, &jsonQuery.Entities, jsonQuery.Limit)
		} else if relation.EntityTo >= 0 {
			// if there is only a to-node
			toName := fmt.Sprintf("n%v", relation.EntityTo)

			ret += *createNodeLet(&jsonQuery.Entities[relation.EntityTo], &toName)

			ret += *createRelationLetWithOnlyToEntity(&relation, relationName, &jsonQuery.Entities, jsonQuery.Limit)
			// Add this relation to the list
		} else {
			fmt.Println("Relation-only queries are currently not supported")
			continue
		}

		// Add this relation to the list
		relationsToReturn = append(relationsToReturn, relationName)
	}

	// Add node let statements for nodes that are not yet returned
	// Create a set from all the entity-from's and entity-to's, to check if they are returned
	nodeSet := make(map[int]bool)
	for _, relation := range jsonQuery.Relations {
		nodeSet[relation.EntityFrom] = true
		nodeSet[relation.EntityTo] = true
	}

	// Check if the entities to return are already returned
	for _, entityIndex := range jsonQuery.Return.Entities {
		if !nodeSet[entityIndex] {
			// If not, return this node
			name := fmt.Sprintf("n%v", entityIndex)
			ret += *createNodeLet(&jsonQuery.Entities[entityIndex], &name)

			// Add this node to the list
			nodesToReturn = append(nodesToReturn, name)
		}
	}

	// Create UNION statements that create unique lists of all the nodes and relations
	// Thus removing all duplicates
	nodeUnion = "\nLET nodes = first(RETURN UNION_DISTINCT("
	for _, relation := range relationsToReturn {
		nodeUnion += fmt.Sprintf("flatten(%v[**].vertices), ", relation)
	}

	for _, node := range nodesToReturn {
		nodeUnion += fmt.Sprintf("%v,", node)
	}
	nodeUnion += "[],[]))\n"

	relationUnion = "LET edges = first(RETURN UNION_DISTINCT("
	for _, relation := range relationsToReturn {
		relationUnion += fmt.Sprintf("flatten(%v[**].edges), ", relation)
	}
	relationUnion += "[],[]))\n"

	ret += nodeUnion + relationUnion
	ret += "RETURN {\"vertices\":nodes, \"edges\":edges }"

	return &ret
}

/* createNodeLet generates a 'LET' statement for a node related query
Parameters: node is an entityStruct containing the information of a single node,
name is the autogenerated name of the node consisting of "n" + the index of the node

Return: a string containing a single LET-statement in AQL
*/
func createNodeLet(node *entity.QueryEntityStruct, name *string) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN %v \n", *name, node.Type)
	footer := "\tRETURN x\n)\n"
	constraints := *createConstraintStatements(&node.Constraints, "x", false)

	ret := header + constraints + footer
	return &ret
}

/* createRelationLetWithFromEntity generates a 'LET' statement for relations with an 'EntityFrom' property and optionally an 'EntitiyTo' property
Parameters: relation is a relation struct containing the information of a single relation,
name is the autogenerated name of the node consisting of "r" + the index of the relation,
entities is a list of entityStructs that are needed to form the relation LET-statement

Return: a string containing a single LET-statement in AQL
*/
func createRelationLetWithFromEntity(relation *entity.QueryRelationStruct, name string, entities *[]entity.QueryEntityStruct, limit int) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", name, relation.EntityFrom)
	forStatement := fmt.Sprintf("\tFOR v, e, p IN %v..%v OUTBOUND x %s \n", relation.Depth.Min, relation.Depth.Max, relation.Type)

	// Guarantees that there is no path returned with a duplicate edge
	// This way there are no cycle paths possible, TODO: more research about this needed
	optionStmtn := "\tOPTIONS { uniqueEdges: \"path\" }\n"

	vFilterStmnt := ""
	if relation.EntityTo != -1 {
		// If there is a to-node, generate the filter statement
		toConstraints := (*entities)[relation.EntityTo].Constraints
		vFilterStmnt += *createConstraintStatements(&toConstraints, "v", false)

		// Add a WITH statement if the collection of entityTo is not yet included
		if (*entities)[(*relation).EntityFrom].Type != (*entities)[(*relation).EntityTo].Type {
			header = fmt.Sprintf("WITH %v\n %v", (*entities)[(*relation).EntityTo].Type, header)
		}
	}

	relationFilterStmnt := *createConstraintStatements(&relation.Constraints, "p", true)
	footer := fmt.Sprintf("\tLIMIT %v \nRETURN DISTINCT p )\n", limit)

	ret := header + forStatement + optionStmtn + vFilterStmnt + relationFilterStmnt + footer
	return &ret
}

/* createRelationLetWithOnlyToEntity generates a 'LET' statement for relations with only an 'EntityTo' property
Parameters: relation is a relation struct containing the information of a single relation,
name is the autogenerated name of the node consisting of "r" + the index of the relation,
entities is a list of entityStructs that are needed to form the relation LET-statement

Return: a string containing a single LET-statement in AQL
*/
func createRelationLetWithOnlyToEntity(relation *entity.QueryRelationStruct, name string, entities *[]entity.QueryEntityStruct, limit int) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", name, relation.EntityTo)
	forStatement := fmt.Sprintf("\tFOR v, e, p IN %v..%v INBOUND x %s \n", relation.Depth.Min, relation.Depth.Max, relation.Type)

	// Guarantees that there is no path returned with a duplicate edge
	// This way there are no cycle paths possible, TODO: more research about this needed
	optionStmtn := "\tOPTIONS { uniqueEdges: \"path\" }\n"

	relationFilterStmnt := *createConstraintStatements(&relation.Constraints, "p", true)
	footer := fmt.Sprintf("\tLIMIT %v \nRETURN DISTINCT p )\n", limit)

	ret := header + forStatement + optionStmtn + relationFilterStmnt + footer
	return &ret
}

// WIP TODO: nodes worden nog niet meegereturned, ff googlen hoe en wat
// func createEntitylessRelationLet(edge *relationStruct, name *string, onlyEdge bool) *string {
// 	var (
// 		header        string
// 		forEdge       string
// 		forSecondNode string
// 	)

// 	footer := "\n\tLIMIT 100\n\tRETURN { vertices: p.vertices[*], edges: p.edges[*] }\n)\n"

// 	// WICKED SWITCHES LETSAGO
// 	if edge.EntityFrom != -1 {
// 		// # 1 (2) Outbound only
// 		header = fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", *name, edge.EntityFrom)
// 		forEdge = fmt.Sprintf("\tFOR v, e, p IN %v..%v OUTBOUND x %s \n", edge.Depth.Min, edge.Depth.Max, edge.Type)
// 		if edge.EntityTo != -1 {
// 			// # 2 Node Relation Node
// 			forSecondNode = fmt.Sprintf("\tFILTER v IN n%v \n", edge.EntityTo)
// 		}
// 	} else {
// 		if edge.EntityTo != -1 {
// 			// # 3 Inbound only
// 			header = fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", *name, edge.EntityTo)
// 			forEdge = fmt.Sprintf("\tFOR v, e, p IN %v..%v INBOUND x %s \n", edge.Depth.Min, edge.Depth.Max, edge.Type)
// 		} else {
// 			// # 4 Relation with constraints only
// 			header = fmt.Sprintf("LET %v = (\n\tFOR x IN %v \n", *name, edge.Type)
// 			footer = "\tLIMIT 100\n\tRETURN x\n)\n"
// 		}
// 	}
// 	allStatement := true
// 	if onlyEdge {
// 		allStatement = false
// 	}

// 	constraints := *createConstraintStatements(&edge.Constraints, *name, allStatement)

// 	ret := header + forEdge + forSecondNode + constraints + footer
// 	return &ret
// }

/*
#1
{
  "Return": {
    "Entities": [
      0,
      1
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    },
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": 0,
      "EntityTo": 1,
      "Constraints": []
    }
  ]
}
*/

/*
#2

{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": 0,
      "EntityTo": -1,
      "Constraints": []
    }
  ]
}
*/

/*
#3
{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": -1,
      "EntityTo": 0,
      "Constraints": []
    }
  ]
}
*/

/*
#4
{
  "Return": {
    "Entities": [],
    "Relations": [
      0
    ]
  },
  "Entities": [],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": -1,
      "EntityTo": -1,
      "Constraints": [
        {
          "Attribute": "Month",
          "Value": "1",
          "DataType": "number",
          "MatchType": "EQ"
        },
        {
          "Attribute": "Day",
          "Value": "15",
          "DataType": "number",
          "MatchType": "EQ"
        }
      ]
    }
  ]
}
*/

/*
#5
{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": []
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": [
        {
          "Attribute": "city",
          "Value": "New York",
          "DataType": "text",
          "MatchType": "exact"
        },
        {
          "Attribute": "country",
          "Value": "USA",
          "DataType": "text",
          "MatchType": "exact"
        }
      ]
    }
  ],
  "Relations": []
}
*/
